request = b"""NOTIFY * HTTP/1.1
Host: 239.255.255.250:1982
Cache-Control: max-age=3600
Location: yeelight://192.168.1.239:55443
NTS: ssdp:alive
Server: POSIX, UPnP/1.0 YGLC/1
id: 0x000000000015243f
model: color
fw_ver: 18
support: get_prop set_default set_power toggle set_bright start_cf stop_cf set_scene cron_add cron_get cron_del set_ct_abx set_rgb
power: on
bright: 100
color_mode: 2
ct: 4000
rgb: 16711680
hue: 100
sat: 35
name: my_bulb""".replace(b'\n', b'\r\n')

response = b"""HTTP/1.1 200 OK
Cache-Control: max-age=3600
Date:
Ext:
Location: yeelight://192.168.1.239:55443
Server: POSIX UPnP/1.0 YGLC/1
id: 0x000000000015243f
model: color
fw_ver: 18
support: get_prop set_default set_power toggle set_bright start_cf stop_cf set_scene cron_add cron_get cron_del set_ct_abx set_rgb
power: on
bright: 100
color_mode: 2
ct: 4000
rgb: 16711680
hue: 100
sat: 35
name: my_bulb""".replace(b'\n', b'\r\n')

response_wrong_location = b"""HTTP/1.1 200 OK
Cache-Control: max-age=3600
Date:
Ext:
Location: yeelight://not.an.ip:55443
Server: POSIX UPnP/1.0 YGLC/1""".replace(b'\n', b'\r\n')



import asyncio
import json
import logging
import socket
import time

from yeelib.exceptions import YeelightError

__all__ = ('Bulb', )

logger = logging.getLogger('yeelib')


class Bulb:
    def __init__(self, ip, port=55443, status_refresh_interval=3600, _registry=None, **kwargs):
        self._registry = _registry
        self.ip = ip
        self.port = port
        self.status_refresh_interval = status_refresh_interval

        support = kwargs.pop('support', [])
        if isinstance(support, str):
            support = support.split()
        self.support = support

        for key, value in kwargs.items():
            setattr(self, key, value)

        self.last_seen = time.time()
        self.__command_id = 0
        self.__socket = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.socket = None

    def __repr__(self):
        return "<%s: %s>" % (type(self).__name__, self.ip)

    @property
    async def socket(self):
        if self.__socket is None:
            self.__socket = await asyncio.open_connection(
                self.ip, self.port)
        return self.__socket

    @socket.setter
    def socket(self, value):
        try:
            reader, writer = self.__socket
        except TypeError:
            pass
        else:
            writer.close()
            if self._registry:
                del self._registry[self.id]
        self.__socket = value

    async def send_command(self, method, params):
        if method not in self.support:
            msg = "The method '%s' is not supported by this bulb." % method
            raise YeelightError(msg)
        self.__command_id += 1
        data = {
            'id': self.__command_id,
            'method': method,
            'params': params,
        }
        msg = json.dumps(data) + '\r\n'
        reader, writer = await self.socket
        try:
            writer.write(msg.encode())
            logger.debug("%s: >>> %s", self.ip, msg.strip())
        except socket.error:
            logger.exception("Connection error")
            self.socket = None
        else:
            try:
                response = (await reader.readline()).decode()
            except OSError:
                logger.exception("Connection error")
                self.socket = None
            else:
                if response is not None:
                    logger.debug("%s: ... %s",
                                 self.ip, response.strip())
                    try:
                        return json.loads(response)
                    except ValueError:
                        logger.exception("Could not read message: %s", response)



import asyncio
import fcntl
import logging
import os
import re
import socket
import struct
import time

from ssdp import SSDPRequest, SimpleServiceDiscoveryProtocol

from .devices import Bulb
from .exceptions import YeelightError

__all__ = ('search_bulbs', 'YeelightProtocol', 'bulbs')

logger = logging.getLogger('yeelib')

bulbs = {}

MCAST_PORT = 1982
MCAST_ADDR = SimpleServiceDiscoveryProtocol.MULTICAST_ADDRESS, MCAST_PORT


async def send_search_broadcast(transport, search_interval=30):
    request = SSDPRequest('M-SEARCH', headers=[
        ('HOST', '%s:%s' % MCAST_ADDR),
        ('MAN', '"ssdp:discover"'),
        ('ST', 'wifi_bulb'),
    ])
    while True:
        try:
            request.sendto(transport, MCAST_ADDR)
        except OSError:
            logger.exception("Connection error")
        await asyncio.sleep(search_interval)


class YeelightProtocol(SimpleServiceDiscoveryProtocol):
    excluded_headers = ['DATE', 'EXT', 'SERVER', 'CACHE-CONTROL', 'LOCATION']
    location_patter = r'yeelight://(?P<ip>\d{1,3}(\.\d{1,3}){3}):(?P<port>\d+)'

    def __init__(self, bulb_class=Bulb, loop=None, **kwargs):
        self.bulb_class = bulb_class
        self.bulb_kwargs = kwargs
        self.loop = loop or asyncio.get_event_loop()

    def connection_made(self, transport):
        self.transport = transport
        ucast_socket = transport.get_extra_info('socket')
        try:
            ucast_socket.bind(('', MCAST_PORT))
            fcntl.fcntl(ucast_socket, fcntl.F_SETFL, os.O_NONBLOCK)
            group = socket.inet_aton(
                SimpleServiceDiscoveryProtocol.MULTICAST_ADDRESS)
            mreq = struct.pack("4sl", group, socket.INADDR_ANY)
            ucast_socket.setsockopt(
                socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
        except socket.error as e:
            ucast_socket.close()
            self.connection_lost(exc=e)
        else:
            self.bcast_msg_task = asyncio.Task(send_search_broadcast(transport))

    @classmethod
    def header_to_kwargs(cls, headers):
        headers = dict(headers)
        location = headers.get('Location')
        cache_control = headers.get('Cache-Control', 'max-age=3600')
        headers = {
            k: v for k, v in headers.items()
            if k.upper() not in cls.excluded_headers
        }

        match = re.match(cls.location_patter, location)
        if match is None:
            raise YeelightError('Location does not match: %s' % location)
        ip = match.groupdict()['ip']
        port = match.groupdict()['port']

        max_age = int(cache_control.rsplit('=', 1)[-1])

        kwargs = dict(ip=ip, port=port, status_refresh_interv=max_age)
        kwargs.update(headers)
        return kwargs

    def request_received(self, request, addr):
        if request.method == 'M-SEARCH':
            return
        self.register_bulb(**self.header_to_kwargs(request.headers))

    def response_received(self, response, addr):
        self.register_bulb(**self.header_to_kwargs(response.headers))

    def register_bulb(self, **kwargs):
        idx = kwargs['id']
        if idx not in bulbs:
            bulbs[idx] = self.bulb_class(**kwargs, **self.bulb_kwargs, _registry=bulbs)
        else:
            bulbs[idx].last_seen = time.time()

    def connection_lost(self, exc):
        logger.exception("connection error")
        self.bcast_msg_task.cancel()
        self.transport.close()

        async def _restart():
            await asyncio.sleep(10)
            await search_bulbs(self.bulb_class, self.loop, kwargs=self.bulb_class)

        asyncio.Task(_restart())


async def remove_missing_bulbs(timeout=60):
    while True:
        missing_bulbs = [
            idx
            for idx, bulb in bulbs.items()
            if bulb.last_seen < time.time() - timeout
        ]
        for idx in missing_bulbs:
            logger.info('%r has not been seen in a while and will be removed', bulbs[idx])
            del bulbs[idx]
        await asyncio.sleep(timeout/2)


async def search_bulbs(bulb_class=Bulb, loop=None, kwargs=None):
    if loop is None:
        loop = asyncio.get_event_loop()
    if kwargs is None:
        kwargs = {}
    unicast_connection = loop.create_datagram_endpoint(
        lambda: YeelightProtocol(bulb_class, **kwargs), family=socket.AF_INET)
    ucast_transport, _ = await unicast_connection
    asyncio.Task(remove_missing_bulbs())
    return bulbs



import asyncio

import pytest

from yeelib.discover import YeelightProtocol, search_bulbs
from yeelib.exceptions import YeelightError
from yeelib import bulbs

from . import fixtures


class TestYeelightProtocoll:
    def test_notify(self, ):
        p = YeelightProtocol()
        p.datagram_received(data=fixtures.request, addr=('192.168.1.239', 1982))
        assert len(bulbs) == 1
        assert bulbs['0x000000000015243f'].ip == '192.168.1.239'

    def test_mcast(self, ):
        p = YeelightProtocol()
        p.datagram_received(data=fixtures.response, addr=('192.168.1.239', 1982))
        assert len(bulbs) == 1
        assert bulbs['0x000000000015243f'].ip == '192.168.1.239'

    def test_duplicate(self):
        p = YeelightProtocol()
        p.datagram_received(data=fixtures.request, addr=('192.168.1.239', 1982))
        p.datagram_received(data=fixtures.request, addr=('192.168.1.239', 1982))
        assert len(bulbs) == 1
        assert bulbs['0x000000000015243f'].ip == '192.168.1.239'

    def test_wrong_location(self):
        p = YeelightProtocol()
        with pytest.raises(YeelightError) as e:
            p.datagram_received(data=fixtures.response_wrong_location,
                                addr=('192.168.1.239', 1982))
        assert 'Location does not match: yeelight://not.an.ip:55443' in str(e)


def test_search_bulbs():
    asyncio.Task(search_bulbs())
    loop = asyncio.get_event_loop()
    loop.run_until_complete(asyncio.sleep(1))



import asyncio

from mocket import Mocket, MocketEntry, mocketize

from yeelib import Bulb


class TestBulb:
    bulb_addr = ('192.168.1.239', 55443)

    @mocketize
    def test_send_command(self):
        Mocket.register(MocketEntry(self.bulb_addr, [b'{"id":1, "result":["ok"]}\r\n']))
        with Bulb(*self.bulb_addr, id=1, support=['set_ct_abx']) as b:
            loop = asyncio.get_event_loop()
            response = loop.run_until_complete(b.send_command('set_ct_abx', [3500, 'smooth', 500]))

        assert response == {'id': 1, 'result': ['ok']}

    def test_repr(self):
        with Bulb(*self.bulb_addr) as b:
            assert repr(b) == '<Bulb: 192.168.1.239>'

    def test_kwargs(self):
        kwargs = {
            'id': '0x000000000015243f',
            'model': 'color',
            'fw_ver': 18,
            'support': 'get_prop set_default set_power toggle set_bright'
                       ' start_cf stop_cf set_scene cron_add cron_get'
                       ' cron_del set_ct_abx set_rgb',
            'power': 'on',
            'bright': 100,
            'color_mode': 2,
            'ct': 4000,
            'rgb': 16711680,
            'hue': 100,
            'sat': 35,
            'name': 'my_bulb',
        }

        with Bulb(*self.bulb_addr, **kwargs) as b:
            assert b.fw_ver == 18
            assert b.support == ['get_prop', 'set_default', 'set_power', 'toggle', 'set_bright',
                                 'start_cf', 'stop_cf', 'set_scene', 'cron_add', 'cron_get',
                                 'cron_del', 'set_ct_abx', 'set_rgb']
